import logging
import os
import sys
import struct
import time

import numpy

import postinglist

DEFAULT_SIZE = 10000000
DEFAULT_CONSTANT = 2
DEFAULT_NDOCS = 100000000
DEFAULT_GENERATE_OUTPUT = 'postinglist.out'


def execute(command, output_file, a=None, zipf_size=None, n_docs=None, input_file=None):
    if command == 'generate':
        __generate(output_file, a, zipf_size, n_docs)
    elif command == 'encode':
        __encode(output_file, input_file)
    elif command == 'decode':
        __decode(output_file, input_file)


def __generate(output_file, a, zipf_size, n_docs):
    logging.info("Executing: a = %i, zipf_size = %i, n_docs = %i, output = %s", a, zipf_size, n_docs, output_file)
    logging.debug("Generating ZIPF distribution with %i values", zipf_size)
    zipf = numpy.random.zipf(a, zipf_size)
    count = numpy.bincount(zipf)
    bins = numpy.nonzero(count)[0]
    distr = zip(bins, count[bins])

    logging.debug("Writing posting list to file: %s", output_file)
    with open(output_file, 'wb') as list_fp:
        mapping_fp = open('mapping-%s' % output_file, 'wb')
        byte_offset = 0
        for item in distr:
            logging.debug("Processing key %s (%i occurences)", item, item[1])
            doc_list = {}
            start = time.time()
            last_ten = time.time()
            length = 0
            for i in range(min(item[1], n_docs)):
                logging.debug("Occurence #%i of key %s", i, item)
                attempts = 0
                while True and attempts < n_docs:
                    doc_id = str(numpy.random.randint(0, n_docs))
                    if doc_id not in doc_list:
                        logging.debug("Key %s received doc_id %s", item[0], doc_id)
                        length += len(doc_id) + 1
                        doc_list[doc_id] = None
                        break
                    else:
                        attempts += 1
                        logging.debug("Unable to find available doc id after %i attempts", attempts)
                if i % 10000 == 0:
                    now = time.time()
                    logging.info("Processed %i/%i items of key %s in %.2f seconds (last 10k: %.2f)",
                                 i, item[1], item[0], now - start, now - last_ten)
                    last_ten = time.time()
            logging.info("Writing posting list for key %s", item[0])
            doc_list = map(int, doc_list.keys())
            doc_list.sort()
            mapping_fp.write('%10d\n' % byte_offset)
            byte_offset += (length + 11)
            list_fp.write('%10d\t%s\n' % (item[0], ','.join(map(str, doc_list))))
        mapping_fp.write('%10d' % byte_offset)
        mapping_fp.close()
        list_fp.close();
    logging.info("Done!")


def __encode(output_file, input_file):
    logging.info("Encoding data from %s into %s", input_file, output_file)
    mapping = os.path.join(os.path.dirname(output_file), 'mapping-%s' % os.path.basename(output_file))
    encoder = postinglist.Encoder(input_file, output_file, mapping)
    encoder.execute(True)



def __decode(output_file, input_file):
    logging.info("Decoding data from %s into %s", input_file, output_file)
    mapping = os.path.join(os.path.dirname(output_file), 'mapping-%s' % os.path.basename(output_file))
    decoder = postinglist.Decoder(input_file, output_file, mapping)
    decoder.execute(True)


if __name__ == '__main__':
    a = DEFAULT_CONSTANT
    zipf_size = DEFAULT_SIZE
    output_file = DEFAULT_GENERATE_OUTPUT
    input_file = DEFAULT_GENERATE_OUTPUT
    n_docs = DEFAULT_NDOCS
    level = logging.INFO

    if len(sys.argv) > 1:
        command = sys.argv[1]
        if command == 'generate':
            if len(sys.argv) > 2:
                for arg in sys.argv[2:]:
                    option = arg[:2]
                    value = arg[2:]
                    if option == '-a':
                        a = int(value)
                    elif option == '-z':
                        zipf_size = int(value)
                    elif option == '-d':
                        n_docs = int(value)
                    elif option == '-v':
                        level = logging.DEBUG
                    else:
                        print "Invalid option:", option
                        exit(1)
        elif command == 'encode' or command == 'decode':
            output_file = '%s-%s' % (command, output_file)
            if command == 'decode':
                input_file = 'encode-%s' % input_file
            if len(sys.argv) > 2:
                option = sys.argv[:2]
                value = sys.argv[2:]
                if option == '-v':
                    level = logging.DEBUG
                else:
                    print "Invalid option:", option
                    exit(1)
        else:
            print "Invalid command:", command
            exit(1)

        logging.basicConfig(level=level)
        execute(command, output_file, a, zipf_size, n_docs, input_file)
    else:
        print "Invalid usage."






