import struct


class PostingListAction(object):

    in_file = None
    out_file = None
    out_mapping_file = None

    def __init__(self, in_file=None, out_file=None, out_mapping_file=None):
        """
        Init. a new encoder

        :type in_file: str
        :param in_file: Path to file to read from.

        :type out_file: str
        :param out_file: Where to output encoded contents.

        :type out_mapping_file: str
        :param out_mapping_file: Where to output beginnings of lists used to calculate posting list ranges.
        """

        self.in_file = in_file
        self.out_file = out_file
        self.out_mapping_file = out_mapping_file

    def execute(self, delta=False):
        if self.in_file and self.out_file and self.out_mapping_file:
            self._execute(delta)
        else:
            raise RuntimeError("All variables are not set. Missing one of in_file, out_file, out_mapping_file")

    def _execute(self, delta):
        raise NotImplementedError("Not implemented! This method must be implemented and executed by subclasses!")


class Encoder(PostingListAction):

    def __init__(self, in_file=None, out_file=None, out_mapping_file=None):
        super(Encoder, self).__init__(in_file, out_file, out_mapping_file)

    def _execute(self, delta):
        with open(self.in_file, 'r') as in_fp:
            out_fp = open(self.out_file, 'wb')
            out_mapping = '0\n'.rjust(11)
            posting_list_len = 0
            for line in in_fp:
                term_list_pair = line.split('\t')
                posting_list = term_list_pair[1].split(',')
                if delta:
                    posting_list = self.__delta(posting_list)
                posting_list_bytestream = []
                for number in posting_list:
#                    print "Encoding", number
                    encoded_number = self.encode_number(number)
                    posting_list_bytestream.append(encoded_number)
                    posting_list_len += len(encoded_number)
                out_mapping += ('%d\n' % (posting_list_len + 1)).rjust(11)
                out_fp.write('%s\t%s\n' % (term_list_pair[0], ''.join(posting_list_bytestream)))
            out_fp.close()
            out_mapping_fp = open(self.out_mapping_file, 'w')
            out_mapping_fp.write(out_mapping)
            in_fp.close()

    def encode_number(self, number):
        number = int(number)
        bytes = []
        while True:
            bytes.insert(0, number % 128)
            if number < 128:
                break
            number /= 128
        bytes[-1] += 128
        return struct.pack('%dB' % len(bytes), *bytes)

    def __delta(self, numbers):
        deltas = [int(numbers[i]) - int(numbers[i-1]) for i in range(1, len(numbers))]
        deltas.insert(0, numbers[0])
        return deltas


class Decoder(PostingListAction):

    in_file_mapping = None

    def __init__(self, in_file=None, out_file=None, out_file_mapping=None, in_file_mapping=None):
        super(Decoder, self).__init__(in_file, out_file, out_file_mapping)

        self.in_file_mapping = in_file_mapping

    def _execute(self, delta):
        with open(self.in_file, 'r') as in_fp:
            out_fp = open(self.out_file, 'wb')
            out_mapping = '0\n'.rjust(11)
            posting_list_len = 0
            read_data = True
            for line in in_fp:
                term_list_pair = line.split('\t')
                if len(term_list_pair) > 1:
                    if read_data:
                        posting_list = term_list_pair[1]
                        if posting_list == '\n':
                            read_data = False
                            continue
                    else:
                        posting_list = line
                else:
                    read_data = False
                    continue
                read_data = True
                if delta:
                    numbers = self.__sum(self.decode_bytestream(posting_list))
                else:
                    numbers = self.decode_bytestream(posting_list)
                numbers = str(numbers).replace(' ', '')
                numbers = numbers[1:len(numbers) - 1]
                posting_list_len += len(numbers)
                out_mapping += ('%d\n' % (posting_list_len + 1)).rjust(11)
                out_fp.write('%s\t%s' % (term_list_pair[0], numbers))
            out_fp.close()
            out_mapping_fp = open(self.out_mapping_file, 'w')
            out_mapping_fp.write(out_mapping)
            in_fp.close()

    def decode_bytestream(self, bytestream):
        n = 0
        numbers = []
        bytestream = struct.unpack('%dB' % len(bytestream), bytestream)
        for byte in bytestream:
            if byte < 128:
                n = 128 * n + byte
            else:
                n = 128 * n + (byte - 128)
                numbers.append(n)
                n = 0
        return numbers

    def __sum(self, numbers):
        for i in range(1, len(numbers)):
            numbers[i] = numbers[i-1] + numbers[i]
        return numbers
