// Copyright 2013 Atbrox AS
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#import "GroupVarintEncoder.h"

#import "CompressionToolbox.h"

@interface GroupVarintEncoder ()

- (UInt8)keyForNumber:(NSUInteger)number;
- (NSData *)encodeNumber:(NSUInteger)a second:(NSUInteger)b third:(NSUInteger)c fourth:(NSUInteger)d;
- (void)stripLeadingZeros:(UInt8 *)buffer atBufferPosition:(NSUInteger)position forNumber:(UInt32)number withKey:(UInt8)key;

@end

@implementation GroupVarintEncoder

@synthesize encodedDataLength =_encodedDataLength;

- (UInt8)keyForNumber:(NSUInteger)number {
    if (!number) return 0;
    return 3 - ([CompressionToolbox CLZ:number | 1] >> 3);
}

- (void)stripLeadingZeros:(UInt8 *)buffer atBufferPosition:(NSUInteger)position forNumber:(UInt32)number withKey:(UInt8)key {
    
    if (key == 0)
        buffer[position] = number & 0xFF;
    else if (key == 1) {
        buffer[position] = number & 0xFF;
        buffer[position + 1] = (number >> 8) & 0xFF;
    }
    else if (key == 2) {
        buffer[position] = number & 0xFF;
        buffer[position + 1] = (number >> 8) & 0xFF;
        buffer[position + 2] = (number >> 16) & 0xFF;
    }
    else {
        buffer[position] = number & 0xFF;
        buffer[position + 1] = (number >> 8) & 0xFF;
        buffer[position + 2] = (number >> 16) & 0xFF;
        buffer[position + 3] = (number >> 24) & 0xFF;
    }
}

- (UInt8)reverseBits:(UInt8)byte {
    return BITREVERSE_TABLE[byte];
}

- (NSUInteger)byteSize:(NSUInteger)number {
    // Key for a number (0, 1, 2 or 3) + 1 is its encoded size.
    
    return [self keyForNumber:number] + 1;
}

- (NSData *)encodeNumber:(NSUInteger)number {
    @throw [NSException exceptionWithName:@"NotImplementedException" reason:@"Method not implemented." userInfo:nil];
}

- (NSData *)encodeNumber:(NSUInteger)a second:(NSUInteger)b third:(NSUInteger)c fourth:(NSUInteger)d {
    
    // Calculate keys (encoding for bit length) for numbers
    
    UInt8 keyForA = [self keyForNumber:a];
    UInt8 keyForB = [self keyForNumber:b];
    UInt8 keyForC = [self keyForNumber:c];
    UInt8 keyForD = [self keyForNumber:d];
    
    UInt8 mask = (keyForD << 6) | (keyForC << 4) | (keyForB << 2) | keyForA;
    
    UInt32 totalLength = keyForA + keyForB + keyForC + keyForD + 5;
    
    UInt8 bits[totalLength];
    UInt8 position = 0;
    bits[position++] = mask;
    [self stripLeadingZeros:bits atBufferPosition:position forNumber:a withKey:keyForA];
    position += keyForA + 1;
    [self stripLeadingZeros:bits atBufferPosition:position forNumber:b withKey:keyForB];
    position += keyForB + 1;
    [self stripLeadingZeros:bits atBufferPosition:position forNumber:c withKey:keyForC];
    position += keyForC + 1;
    [self stripLeadingZeros:bits atBufferPosition:position forNumber:d withKey:keyForD];
    position += keyForD + 1;
    
    return [NSData dataWithBytes:bits length:totalLength];
}

- (NSData *)encode:(CFArrayRef)numbers {
    
    // Look to fix this. Kind of a harsh requirement.
    
//    NSLog(@"Encoding %i numbers...", numbers.count);
    
    NSUInteger itemCount = CFArrayGetCount(numbers);
    
    if (!itemCount || itemCount & 3) {
        NSString *error = [NSString stringWithFormat:@"Array of numbers to encode must be a multiple of 4. Size was %i.", itemCount];
        @throw [NSException exceptionWithName:@"InvalidSizeException" reason:error userInfo:nil];
    }
    
    // Precalculate total space required.
    NSUInteger totalLength = 0;
    for (NSUInteger i = 0; i < itemCount; i++) {
        totalLength += [self byteSize:(NSUInteger)CFArrayGetValueAtIndex(numbers, i)];
    }
    // Plus 1 byte for each block of four integers.
    
    totalLength += itemCount >> 2;
    
    NSMutableData *data = [NSMutableData dataWithLength:totalLength];
    NSUInteger currentPosition = 0;
    for (NSUInteger i = 0; i < itemCount - 3; i += 4) {
        NSUInteger a = (NSUInteger)CFArrayGetValueAtIndex(numbers, i);
        NSUInteger b = (NSUInteger)CFArrayGetValueAtIndex(numbers, i + 1);
        NSUInteger c = (NSUInteger)CFArrayGetValueAtIndex(numbers, i + 2);
        NSUInteger d = (NSUInteger)CFArrayGetValueAtIndex(numbers, i + 3);
        NSData *singleData = [self encodeNumber:a second:b third:c fourth:d];
        [data replaceBytesInRange:NSMakeRange(currentPosition, singleData.length) withBytes:singleData.bytes];
        currentPosition += singleData.length;
        
    }
    
    return data;
}

- (NSString *)scheme {
    return @"Group Varint";
}


@end
