// Copyright 2013 Atbrox AS
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#import "ParallelDecodingModel.h"

#import "Debug.h"
#import "EliasGammaDecoder.h"
#import "VariableByteDecoder.h"
#import "ParallelEncodingModel.h"

@interface ParallelDecodingModel ()

- (void)prefixSum:(CFMutableArrayRef *)deltas;
- (NSString *)joinNumbersIntoString:(CFMutableArrayRef *)numbers forPostingListEntry:(PostingListEntry *)entry;

@end

@implementation ParallelDecodingModel

@synthesize numberOfThreads = _numberOfThreads;

- (id)init {
    self = [super init];
    if (self)
        _decoder = [[PostingListDecoder alloc] init];
    return self;
}

-(void) prefixSum:(CFMutableArrayRef *)deltas
{
    for (NSUInteger i = 1; i < CFArrayGetCount(*deltas); i++)
    {
        NSUInteger current = (NSUInteger)CFArrayGetValueAtIndex(*deltas, i);
        NSUInteger previous = (NSUInteger)CFArrayGetValueAtIndex(*deltas, i - 1);
        NSUInteger sum = current + previous;
        //        NSLog(@"Storing sum: %i + %i = %i at index %i", previous, current, sum, i);
        CFArraySetValueAtIndex(*deltas, i, (void *)sum);
    }
}

- (NSString *)joinNumbersIntoString:(CFMutableArrayRef *)numbers forPostingListEntry:(PostingListEntry *)entry {
    
    NSUInteger numberCount = CFArrayGetCount(*numbers);
    NSMutableString *encodeString = [NSMutableString stringWithCapacity:numberCount << 1];
    
    if (!entry.dataLength) {
        [encodeString appendFormat:@"%@\t", entry.term];
    }
    
    if (_numberOfThreads > 1) {
        
        dispatch_queue_t joinQueue = dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0);
        dispatch_group_t joinGroup = dispatch_group_create();
        
        NSMutableArray *subStrings = [NSMutableArray array];
        for (NSUInteger i = 0; i < _numberOfThreads; i++) {
            [subStrings addObject:[[NSObject alloc] init]];
        }
        
        NSUInteger itemsPerThread = numberCount / _numberOfThreads;
        NSUInteger remainder = numberCount % _numberOfThreads;
        
        for (NSUInteger i = 0; i < _numberOfThreads; i++) {
            
            dispatch_group_async(joinGroup, joinQueue, ^{
                
                NSUInteger start = i * itemsPerThread;
                NSUInteger end = start + itemsPerThread;
                if (i == _numberOfThreads - 1) {
                    end += remainder;
                }
                
                NSMutableString *threadString = [[NSMutableString alloc] init];
                for (NSUInteger j = start; j < end - 1; j++) {
                    [threadString appendFormat:@"%u,", (NSUInteger)CFArrayGetValueAtIndex(*numbers, j)];
                }
                [threadString appendFormat:@"%u", (NSUInteger)CFArrayGetValueAtIndex(*numbers, end -1)];
                //            NSArray *threadArray = [numbers subarrayWithRange:NSMakeRange(start, end - start)];
                @synchronized(subStrings) {
                    subStrings[i] = threadString;
                }
            });
        }
        dispatch_group_wait(joinGroup, DISPATCH_TIME_FOREVER);
        
        for (NSString *subString in subStrings) {
            [encodeString appendFormat:@"%@,", subString];
        }
        
        if (entry.lastReadPosition >= entry.range.location + entry.range.length) {
            [encodeString replaceCharactersInRange:NSMakeRange(encodeString.length - 1, 1) withString:@"\n"];
        }
        
    }
    else {
        
        for (NSUInteger i = 0; i < numberCount - 1; i++) {
            [encodeString appendFormat:@"%u,", (NSUInteger)CFArrayGetValueAtIndex(*numbers, i)];
        }
        [encodeString appendFormat:@"%u", (NSUInteger)CFArrayGetValueAtIndex(*numbers, numberCount - 1)];
        
        if (entry.lastReadPosition < entry.range.location + entry.range.length) {
            [encodeString appendString:@","];
        }
        else {
            [encodeString appendString:@"\n"];
        }
        
    }
    //    NSLog(@"Calculated encode string: %@", encodeString);
    return encodeString;
}

-(BOOL) _preprocess:(PostingListEntry *)preProcessObject
{
    //    sleep(1);
    return YES;
}

-(BOOL) _process:(PostingListEntry *)processObject
{
    //    @autoreleasepool {
    NSUInteger bufferLength = processObject.buffer.length;
    
    if ([_decoder.decoder isKindOfClass:[EliasGammaDecoder class]]) {
        //            NSLog(@"Converting to byte array...");
        bufferLength = bufferLength << 3;
        processObject.buffer = [CompressionToolbox convertToBytes:processObject.buffer from:0 length:bufferLength];
        //            NSLog(@"Done! Length: %i", processObject.buffer.length);
    }
    
    if (processObject.bufferTail) {
        //            NSLog(@"Found buffer tail of length %i: %@. Merging...", processObject.previousBufferTail.length, processObject.previousBufferTail);
        NSMutableData *mutableBufferTail = [NSMutableData dataWithData:processObject.bufferTail];
        [mutableBufferTail appendData:processObject.buffer];
        processObject.buffer = mutableBufferTail;
    }
    
    _decoder.input = processObject.buffer;
    processObject.haveDecodedBuffer = [_decoder decode];
    if (processObject.buffer.length > _decoder.decodeLength) {
        
        // Class specific ranges when setting the tail buffer
        
        NSUInteger from = _decoder.decodeLength;
        NSUInteger length = processObject.buffer.length - _decoder.decodeLength;
        if ([_decoder.decoder isKindOfClass:[VariableByteDecoder class]]) {
            from += 1;
            length -= 1;
        }
        processObject.bufferTail = [processObject.buffer subdataWithRange:NSMakeRange(from, length)];
    }
    else {
        //            NSLog(@"Decode complete at position %i! Erasing buffer tail...", processObject.lastReadPosition);
        processObject.bufferTail = nil;
    }
    
    //    }
    return YES;
}

-(BOOL) _postprocess:(PostingListEntry *)postProcessObject
{
    if (CFArrayGetCount(postProcessObject.haveDecodedBuffer) < 1) {
        return YES;
    }
    @autoreleasepool {
        //        NSMutableArray *tmpArray = [NSMutableArray arrayWithArray:postProcessObject.toEncodeBuffer];
        //        NSLog(@"This run's first value: %@", postProcessObject.encodeBuffer[0]);
        CFMutableArrayRef haveDecodedBuffer = postProcessObject.haveDecodedBuffer;
        if (postProcessObject.lastReadValue) {
            //            NSLog(@"Previous decoded value: %i", postProcessObject.lastReadValue);
            NSUInteger newPrefix = postProcessObject.lastReadValue + (NSUInteger)CFArrayGetValueAtIndex(haveDecodedBuffer, 0);
            CFArraySetValueAtIndex(postProcessObject.haveDecodedBuffer, 0, (void *)newPrefix);
            [self prefixSum:&haveDecodedBuffer];
        }
        else
            [self prefixSum:&haveDecodedBuffer];
        
        postProcessObject.lastReadValue = (NSUInteger)CFArrayGetValueAtIndex(haveDecodedBuffer, CFArrayGetCount(haveDecodedBuffer) - 1);
        //        NSLog(@"Last read value: %i", postProcessObject.lastReadValue);
        
        //        NSLog(@"Decode array: %@", tmpArray);
        
        NSString *encodeString = [self joinNumbersIntoString:&haveDecodedBuffer forPostingListEntry:postProcessObject];
        
        CFRelease(haveDecodedBuffer);
        
        postProcessObject.data = [encodeString dataUsingEncoding:NSUTF8StringEncoding];
        postProcessObject.dataLength += postProcessObject.data.length;
        //        postProcessObject.numberOfEntries += tmpArray.count;
        //        NSLog(@"Decoded values: %@", decodedArray);
    }
    return YES;
}

- (BOOL)_writePostingList:(PostingListEntry *)writeProcessedObject {
    @autoreleasepool {
        if (super.writeDataFileHandle) {
            [super.writeDataFileHandle writeData:writeProcessedObject.data];
        }
    }
    return YES;
}

- (BOOL)_writeMapping:(PostingListEntry *)writeMappingObject {
    @autoreleasepool {
        if (super.writeMappingFileHandle) {
            [super.writeMappingFileHandle writeData:[@"0000000000\n" dataUsingEncoding:NSUTF8StringEncoding]];
            [super.writeMappingFileHandle writeData:[[NSString stringWithFormat:@"%010d", writeMappingObject.dataLength] dataUsingEncoding:NSUTF8StringEncoding]];
            [super.writeMappingFileHandle closeFile];
        }
        NSLog(@"Number of entries: %i", writeMappingObject.numberOfEntries);
    }
    return YES;
}

- (void)_prepareWrite {
    
    NSFileManager *fm = [NSFileManager defaultManager];
    
    NSDateFormatter *dateFormatter = [[NSDateFormatter alloc] init];
    dateFormatter.dateFormat = @"yyyy-MM-dd_HH-mm-ss";
    NSDate *benchmarkDate = [NSDate date];
    
    NSArray *directories = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
    NSString *dataFilename = [NSString stringWithFormat:@"decode-data-%@.log", [dateFormatter stringFromDate:benchmarkDate]];
    NSString *mappingFilename = [NSString stringWithFormat:@"decode-mapping-%@.log", [dateFormatter stringFromDate:benchmarkDate]];
    NSString *dataPath = [directories[0] stringByAppendingPathComponent:dataFilename];
    NSString *mappingPath = [directories[0] stringByAppendingPathComponent:mappingFilename];
    
    BOOL dataFileCreated;
    if (![fm fileExistsAtPath:dataPath])
        dataFileCreated = [fm createFileAtPath:dataPath contents:nil attributes:nil];
    else
    {
        dataFileCreated = [fm removeItemAtPath:dataPath error:nil];
        dataFileCreated = [fm createFileAtPath:dataPath contents:nil attributes:nil];
    }
    
    BOOL mappingFileCreated;
    if (![fm fileExistsAtPath:mappingPath])
        mappingFileCreated = [fm createFileAtPath:mappingPath contents:nil attributes:nil];
    else
    {
        mappingFileCreated = [fm removeItemAtPath:mappingPath error:nil];
        mappingFileCreated = [fm createFileAtPath:mappingPath contents:nil attributes:nil];
    }
    
    if (mappingFileCreated && dataFileCreated) {
        super.writeDataFileHandle = [NSFileHandle fileHandleForWritingAtPath:dataPath];
        super.writeMappingFileHandle = [NSFileHandle fileHandleForWritingAtPath:mappingPath];
    }
    else
        NSLog(@"Could not open filehandles!");
}

-(NSArray *) decodedValuesForTerm:(NSString *)term
{
    return [super.values objectForKey:term];
}

-(NSUInteger) store:(NSString *)path withMappingAtPath:(NSString *)mappingPath inBlocks:(BOOL)blocked
{
    NSFileManager *fm = [NSFileManager defaultManager];
    
    NSUInteger blockSize = [super blockSize];
    
    BOOL fileCreated;
    if (![fm fileExistsAtPath:path])
        fileCreated = [fm createFileAtPath:path contents:nil attributes:nil];
    else
    {
        fileCreated = [fm removeItemAtPath:path error:nil];
        fileCreated = [fm createFileAtPath:path contents:nil attributes:nil];
    }
    
    BOOL mappingFileCreated;
    if (![fm fileExistsAtPath:mappingPath])
        mappingFileCreated = [fm createFileAtPath:mappingPath contents:nil attributes:nil];
    else
    {
        mappingFileCreated = [fm removeItemAtPath:mappingPath error:nil];
        mappingFileCreated = [fm createFileAtPath:mappingPath contents:nil attributes:nil];
    }
    
    if (fileCreated && mappingFileCreated)
    {
        @autoreleasepool {
            NSFileHandle *writer = [NSFileHandle fileHandleForWritingAtPath:path];
            NSFileHandle *mappingWriter = [NSFileHandle fileHandleForWritingAtPath:mappingPath];
            NSArray *sortedKeys = [[super.values allKeys] sortedArrayUsingSelector:@selector(compare:)];
            [Debug log:@"File '%@' exists. Starting write...", path];
            
            NSMutableData *toWriteData = [NSMutableData data];
            NSMutableData *mappingData = [NSMutableData data];
            NSUInteger postingListOffset = 0;
            for (NSUInteger i = 0; i < sortedKeys.count; i++)
            {
                NSString *term = sortedKeys[i];
                
                NSData *termMappingData = [[NSString stringWithFormat:@"%08i\n", postingListOffset] dataUsingEncoding:NSASCIIStringEncoding];
                [mappingData appendData:termMappingData];
                
                NSData *prefix = [[NSString stringWithFormat:@"%@\t", term] dataUsingEncoding:NSUTF8StringEncoding];
                postingListOffset += prefix.length;
                [toWriteData appendData:prefix];
                
                NSArray *postingList = [super.values objectForKey:term];
                [Debug log:@"Processing term '%@'. Size: %i.", term, postingList.count];
                for (NSUInteger j = 0; j < postingList.count; j++)
                {
                    if (blocked)
                    {
                        if (toWriteData.length >= blockSize)
                        {
                            [writer writeData:toWriteData];
                            [toWriteData setLength:0];
                        }
                    }
                    NSNumber *number = postingList[j];
                    NSData *data;
                    if (j < postingList.count - 1)
                        data = [[NSString stringWithFormat:@"%@,", number] dataUsingEncoding:NSASCIIStringEncoding];
                    else
                        data = [[NSString stringWithFormat:@"%@", number] dataUsingEncoding:NSASCIIStringEncoding];
                    postingListOffset += data.length;
                    [toWriteData appendData:data];
                }
                
                if (i != sortedKeys.count - 1)
                {
                    [toWriteData appendData:[@"\n" dataUsingEncoding:NSUTF8StringEncoding]];
                    postingListOffset += 1;
                }
                if (!blocked)
                {
                    [writer writeData:toWriteData];
                    [toWriteData setLength:0];
                }
                [Debug log:@"Processed term '%@'. Written data: %i", term, postingListOffset];
            }
            [writer writeData:toWriteData];
            [writer closeFile];
            
            [mappingData appendData:[[NSString stringWithFormat:@"%08i", postingListOffset + 1] dataUsingEncoding:NSASCIIStringEncoding]];
            BOOL mappingFileCreated;
            if (![fm fileExistsAtPath:mappingPath])
                mappingFileCreated = [fm createFileAtPath:mappingPath contents:nil attributes:nil];
            else
                mappingFileCreated = YES;
            if (mappingFileCreated)
            {
                [Debug log:@"Mapping file '%@' exists. Writing mapping file...", mappingPath];
                [mappingWriter writeData:mappingData];
                [mappingWriter closeFile];
            }
        }
    }
    else
    {
        [Debug log:@"Unable to find/create file '%@'.", path];
        return -1;
    }
    
    return 0;
}

@end
