// Copyright 2013 Atbrox AS
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#import "ParallelEncodingModel.h"

#import "Debug.h"
#import "EliasGammaEncoder.h"
#import "GroupVarintEncoder.h"
#import "PostingListEncoder.h"

#define DELTA_ENCODE 1

@interface ParallelEncodingModel ()

- (NSArray *)partitionElementsAmongThreads:(CFMutableArrayRef)entries;

@end

@implementation ParallelEncodingModel {
    NSArray *threadPartitioning;
}

@synthesize numberOfThreads = _numberOfThreads;

- (id)init {
    self = [super init];
    if (self)
        _encoder = [[PostingListEncoder alloc] init];
    return self;
}

-(NSDictionary *) encodedUrlsForTerm:(NSString *)term
{
    return [super.values objectForKey:term];
}

-(BOOL) _preprocess:(PostingListEntry *)preProcessObject
{
    //    @autoreleasepool {
    [Debug log:@"Processing read for term '%@' in range (Location: %i, Length: %i)",
     preProcessObject.term,
     preProcessObject.range.location,
     preProcessObject.range.length];
    
    // Join with data from previous read.
    
    NSString *termDataAsString = [[NSString alloc] initWithData:preProcessObject.buffer encoding:NSASCIIStringEncoding];
    
    if (preProcessObject.bufferTail) {
        termDataAsString = [[[NSString alloc] initWithData:preProcessObject.bufferTail encoding:NSASCIIStringEncoding] stringByAppendingString:termDataAsString];
    }
    
    // Join with remainder from previous encode.
    
//    NSMutableArray *termPostingList;
    CFMutableArrayRef termPostingList;
    if (preProcessObject.toEncodeBufferTail) {
//        NSLog(@"Found encode buffer tail: %@", preProcessObject.toEncodeBufferTail);
        termPostingList = CFArrayCreateMutableCopy(kCFAllocatorDefault, 0, preProcessObject.toEncodeBufferTail);
    }
    else {
        termPostingList = CFArrayCreateMutable(kCFAllocatorDefault, 0, NULL);
    }
    
    NSMutableString *tokenizerString = [termDataAsString mutableCopy];
    
    char *postingListToken = strtok((char *)[tokenizerString UTF8String], ",");
    while (postingListToken != NULL) {
        CFArrayAppendValue(termPostingList, (void *)strtol(postingListToken, NULL, 0));
        postingListToken = strtok(NULL, ",");
    }
    
    NSUInteger termPostingListCount = CFArrayGetCount(termPostingList);
    
    if ([termDataAsString characterAtIndex:termDataAsString.length - 1] == ',') {
        preProcessObject.bufferTail = 0;
        
        // Ending with comma means an extra empty entry in termPostingList
        
        if (CFArrayGetValueAtIndex(termPostingList, termPostingListCount - 1) == 0) {
            CFArrayRemoveValueAtIndex(termPostingList, termPostingListCount - 1);
        }
    }
    else {
        NSRange lastOccurenceOfComma = [termDataAsString rangeOfString:@"," options:NSBackwardsSearch];
        if (lastOccurenceOfComma.location == NSNotFound) {
            preProcessObject.bufferTail = 0;
        }
        else {
            preProcessObject.bufferTail = [[termDataAsString substringFromIndex:lastOccurenceOfComma.location + 1] dataUsingEncoding:NSASCIIStringEncoding];
        }
    }
    
    termPostingListCount = CFArrayGetCount(termPostingList);
    NSUInteger encodeRange;
    
    if (!preProcessObject.bufferTail || preProcessObject.lastReadPosition >= preProcessObject.range.location + preProcessObject.range.length)
        encodeRange = termPostingListCount;
    else
        encodeRange = termPostingListCount - 1;
    
    if (encodeRange < termPostingListCount) {
        CFArrayRemoveValueAtIndex(termPostingList, termPostingListCount - 1);
    }
    termPostingListCount = CFArrayGetCount(termPostingList);
    encodeRange = termPostingListCount;
    
    if (_numberOfThreads > 1) {
        
        if ([_encoder.encoder isKindOfClass:[GroupVarintEncoder class]]) {
            
            // If the last encoded values aren't divisable by four, append zeros to make it.
            
            if (preProcessObject.lastReadPosition >= preProcessObject.range.location + preProcessObject.range.length) {
                NSUInteger missingItemsCount = (termPostingListCount & 3);
//                NSLog(@"Adding %i zeros...", missingItemsCount);
                for (NSUInteger i = 0; i < missingItemsCount; i++) {
                    CFArrayAppendValue(termPostingList, 0);
                }
            }
        }
        
        threadPartitioning = [self partitionElementsAmongThreads:termPostingList];
        
        ThreadEntryCount lastEntry;
        [threadPartitioning[threadPartitioning.count - 1] getValue:&lastEntry];
//        NSLog(@"Last entry: Start: %i. End: %i", lastEntry.start, lastEntry.end);
        termPostingListCount = CFArrayGetCount(termPostingList);
        if (lastEntry.end < termPostingListCount) {
            
            // Set tail of this buffer
            
            if (preProcessObject.toEncodeBufferTail) {
                CFRelease(preProcessObject.toEncodeBufferTail);
            }
            
            NSUInteger bufferTailLength = termPostingListCount - lastEntry.end;
            NSUInteger *bufferTail = (NSUInteger *)malloc(sizeof(NSUInteger) * bufferTailLength);
            CFArrayGetValues(termPostingList, CFRangeMake(lastEntry.end, bufferTailLength), (const void **)bufferTail);
            CFArrayRef temporaryBufferTail = CFArrayCreate(kCFAllocatorDefault, (const void **)bufferTail, bufferTailLength, NULL);
            preProcessObject.toEncodeBufferTail = CFArrayCreateMutableCopy(kCFAllocatorDefault, 0, temporaryBufferTail);
            CFRelease(temporaryBufferTail);
            free(bufferTail);
            
            // Remove tail from this buffer
            
            NSUInteger numEntries = termPostingListCount - bufferTailLength;
            NSUInteger *newPostingList = (NSUInteger *)malloc(sizeof(NSUInteger) * numEntries);
            CFArrayGetValues(termPostingList, CFRangeMake(0, numEntries), (const void **)newPostingList);
            CFArrayRef temporaryBuffer = CFArrayCreate(kCFAllocatorDefault, (const void **)newPostingList, numEntries, NULL);
            termPostingList = CFArrayCreateMutableCopy(kCFAllocatorDefault, 0, temporaryBuffer);
            termPostingListCount = CFArrayGetCount(termPostingList);
            encodeRange = termPostingListCount;
            CFRelease(temporaryBuffer);
            free(newPostingList);
//            NSLog(@"New length of posting list: %i", termPostingList.count);
        }
        else {
            preProcessObject.toEncodeBufferTail = nil;
        }
    }
    else {
        
        // Group Varint requires special care when encoding. Posting list must be divisable by 4.
        
        threadPartitioning = nil;
        if (preProcessObject.lastReadPosition < preProcessObject.range.location + preProcessObject.range.length) {
            if ([_encoder.encoder isKindOfClass:[GroupVarintEncoder class]]) {
                NSUInteger numEntries = termPostingListCount;
                numEntries -= numEntries & 3;
//                NSLog(@"Current entries: %i. Must be reduced to: %i.", termPostingListCount, numEntries);
                
                // Set tail of this buffer
                
                NSUInteger bufferTailLength = termPostingListCount - numEntries;
                NSUInteger *bufferTail = (NSUInteger *)malloc(sizeof(NSUInteger) * bufferTailLength);
                CFArrayGetValues(termPostingList, CFRangeMake(numEntries, bufferTailLength), (const void **)bufferTail);
                CFArrayRef temporaryBufferTail = CFArrayCreate(kCFAllocatorDefault, (const void **)bufferTail, bufferTailLength, NULL);
                preProcessObject.toEncodeBufferTail = CFArrayCreateMutableCopy(kCFAllocatorDefault, 0, temporaryBufferTail);
                CFRelease(temporaryBufferTail);
                free(bufferTail);
                
                // Remove tail from this buffer
                
                NSUInteger *newPostingList = (NSUInteger *)malloc(sizeof(NSUInteger) * numEntries);
                CFArrayGetValues(termPostingList, CFRangeMake(0, numEntries), (const void **)newPostingList);
                CFArrayRef temporaryBuffer = CFArrayCreate(kCFAllocatorDefault, (const void **)newPostingList, numEntries, NULL);
                termPostingList = CFArrayCreateMutableCopy(kCFAllocatorDefault, 0, temporaryBuffer);
                termPostingListCount = CFArrayGetCount(termPostingList);
                encodeRange = termPostingListCount;
                CFRelease(temporaryBuffer);
                free(newPostingList);
//                NSLog(@"To be encoded: %@", termPostingList);
//                NSLog(@"Stored in encode buffer tail: %@", preProcessObject.toEncodeBufferTail);
            }
        }
    }

    
    // First delta is calculated with the last value read last time
    //        NSLog(@"First items in postings list: %i, %i", [termPostingList[0] integerValue], [termPostingList[1] integerValue]);
//    NSLog(@"Encoding: %@", termPostingList);
    NSUInteger currentLastValue = (NSUInteger)CFArrayGetValueAtIndex(termPostingList, encodeRange - 1);
    
    for (NSUInteger j = encodeRange - 1; j > 0; j--)
    {
        NSUInteger larger = (NSUInteger)CFArrayGetValueAtIndex(termPostingList, j);
        NSUInteger lesser = (NSUInteger)CFArrayGetValueAtIndex(termPostingList, j - 1);
        NSUInteger delta = larger - lesser;
        //                            NSLog(@"Calculated delta: %li - %li = %@ inserted at %li", larger, lesser, delta, j);
        CFArraySetValueAtIndex(termPostingList, j, (void *)delta);
    }
    
    if (preProcessObject.lastReadValue) {
//        NSLog(@"Found last read value: %i", preProcessObject.lastReadValue);
        NSUInteger newHeadSum = (NSUInteger)CFArrayGetValueAtIndex(termPostingList, 0) - preProcessObject.lastReadValue;
        CFArraySetValueAtIndex(termPostingList, 0, (void *)newHeadSum);
    }
    
    preProcessObject.lastReadValue = currentLastValue;
//    NSLog(@"Set last read value: %i", preProcessObject.lastReadValue);
    preProcessObject.toEncodeBuffer = termPostingList;
    preProcessObject.numberOfEntries += CFArrayGetCount(preProcessObject.toEncodeBuffer);
    //    }
    return YES;
}

-(BOOL) _process:(PostingListEntry *)processObject
{
//    @autoreleasepool {
    
        _encoder.input = processObject.toEncodeBuffer;
//    NSLog(@"Sending values to threads: %@", _encoder.input);
        processObject.data = [_encoder encode:threadPartitioning];
    CFRelease(processObject.toEncodeBuffer);
//    }
    return YES;
}

-(BOOL) _postprocess:(PostingListEntry *)postProcessObject
{
//    @autoreleasepool {
        // If using Elias Gamma Coding, trailing bits must be removed by replacing them with the current buffer's head.
        
        if ([_encoder.encoder isKindOfClass:[EliasGammaEncoder class]]) {
            
//            NSLog(@"Starting split/merge operation. Encoded %i terms so far.", postProcessObject.numberOfEntries);
            
            if (postProcessObject.encodedDataTail) {
//                NSLog(@"Found previous data of %i bytes: %@. Old length is: %i. Merging...", postProcessObject.lastEncodedData.length, postProcessObject.lastEncodedData, postProcessObject.data.length);
                NSMutableData *extendedData = [NSMutableData dataWithData:postProcessObject.encodedDataTail];
                [extendedData appendData:postProcessObject.data];
                postProcessObject.data = extendedData;
//                NSLog(@"Done merging! New length: %i", postProcessObject.data.length);
            }
            
            if (postProcessObject.lastReadPosition < postProcessObject.range.location + postProcessObject.range.length) {
                
                // Data to write must be divisable by 8 to ensure garbage is not written.
                
                NSUInteger bitCount = postProcessObject.data.length;
                NSUInteger byteAlignedBitCount = bitCount & 7 ? ((bitCount >> 3) + 1) << 3 : bitCount;
                NSUInteger bitTrailCount = byteAlignedBitCount - bitCount;
//                NSLog(@"Bit length: %i. Aligned bit length: %i. Difference: %i", bitCount, byteAlignedBitCount, bitTrailCount);
                if (bitTrailCount) {
                    postProcessObject.encodedDataTail = [postProcessObject.data subdataWithRange:NSMakeRange(bitCount - (8 - bitTrailCount), 8 - bitTrailCount)];
//                    NSLog(@"Storing %i bytes for next write (Bytes %i - %i).",
//                          postProcessObject.lastEncodedData.length,
//                          bitCount - (8 - bitTrailCount),
//                          bitCount - (8 - bitTrailCount) + (8 - bitTrailCount));
//                    NSLog(@"Stored bytes: %@", postProcessObject.lastEncodedData);
//                    NSLog(@"Creating subdata between %i - %i", 0, bitCount - (8 - bitTrailCount));
                    postProcessObject.data = [postProcessObject.data subdataWithRange:NSMakeRange(0, bitCount - (8 - bitTrailCount))];
                }
                else {
                    postProcessObject.encodedDataTail = nil;
                }
//                NSLog(@"Total encoded length: %i", [_encoder.encoder totalEncodedDataLength]);
//                NSLog(@"Bit length: %i", postProcessObject.data.length);
                postProcessObject.data = [CompressionToolbox convertToBits:postProcessObject.data overflow:NO];
            }
            else {
//                NSLog(@"Total encoded length: %i", [_encoder.encoder totalEncodedDataLength]);
//                NSLog(@"Bit length: %i", postProcessObject.data.length);
                postProcessObject.data = [CompressionToolbox convertToBits:postProcessObject.data overflow:YES];
            }
//            NSLog(@"Converting %i bytes to byte array of equal length in bits.", postProcessObject.data.length);
//            NSLog(@"New length: %i bytes.", postProcessObject.data.length);
        }
    if (!postProcessObject.dataLength) {
        NSMutableData *termData = [NSMutableData dataWithData:[[postProcessObject.term stringByAppendingString:@"\t"] dataUsingEncoding:NSUTF8StringEncoding]];
        [termData appendData:postProcessObject.data];
        postProcessObject.data = termData;
    }
//    NSLog(@"Previously written length: %i.", postProcessObject.dataLength);
    postProcessObject.dataLength += postProcessObject.data.length;
//    NSLog(@"New written length: %i\n\n", postProcessObject.dataLength);
//    }
    return YES;
}

- (BOOL)_writePostingList:(PostingListEntry *)writeProcessedObject {
    @autoreleasepool {
        if (super.writeDataFileHandle) {
            [super.writeDataFileHandle writeData:writeProcessedObject.data];
        }
    }
    return YES;
}

- (BOOL)_writeMapping:(PostingListEntry *)writeMappingObject {
    @autoreleasepool {
        if (super.writeMappingFileHandle) {
            [super.writeMappingFileHandle writeData:[@"0000000000\n" dataUsingEncoding:NSUTF8StringEncoding]];
            [super.writeMappingFileHandle writeData:[[NSString stringWithFormat:@"%010d", writeMappingObject.dataLength] dataUsingEncoding:NSUTF8StringEncoding]];
            [super.writeMappingFileHandle closeFile];
        }
        NSLog(@"Number of entries: %i", writeMappingObject.numberOfEntries);
    }
    return YES;
}

- (void)_prepareWrite {
    
    NSFileManager *fm = [NSFileManager defaultManager];
    
    NSDateFormatter *dateFormatter = [[NSDateFormatter alloc] init];
    dateFormatter.dateFormat = @"yyyy-MM-dd_HH-mm-ss";
    NSDate *benchmarkDate = [NSDate date];
    
    NSArray *directories = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
    NSString *dataFilename = [NSString stringWithFormat:@"encode-data-%@.log", [dateFormatter stringFromDate:benchmarkDate]];
    NSString *mappingFilename = [NSString stringWithFormat:@"encode-mapping-%@.log", [dateFormatter stringFromDate:benchmarkDate]];
    NSString *dataPath = [directories[0] stringByAppendingPathComponent:dataFilename];
    NSString *mappingPath = [directories[0] stringByAppendingPathComponent:mappingFilename];
    
    BOOL dataFileCreated;
    if (![fm fileExistsAtPath:dataPath])
        dataFileCreated = [fm createFileAtPath:dataPath contents:nil attributes:nil];
    else
    {
        dataFileCreated = [fm removeItemAtPath:dataPath error:nil];
        dataFileCreated = [fm createFileAtPath:dataPath contents:nil attributes:nil];
    }
    
    BOOL mappingFileCreated;
    if (![fm fileExistsAtPath:mappingPath])
        mappingFileCreated = [fm createFileAtPath:mappingPath contents:nil attributes:nil];
    else
    {
        mappingFileCreated = [fm removeItemAtPath:mappingPath error:nil];
        mappingFileCreated = [fm createFileAtPath:mappingPath contents:nil attributes:nil];
    }
    
    if (mappingFileCreated && dataFileCreated) {
        super.writeDataFileHandle = [NSFileHandle fileHandleForWritingAtPath:dataPath];
        super.writeMappingFileHandle = [NSFileHandle fileHandleForWritingAtPath:mappingPath];
    }
    else
        NSLog(@"Could not open filehandles!");
}

- (NSArray *)partitionElementsAmongThreads:(CFMutableArrayRef)entries {
    NSUInteger numElements = CFArrayGetCount(entries);
//    NSLog(@"Total number of elements: %i", numElements);
    
    NSMutableArray *_threadPartitioning = [NSMutableArray array];
    NSUInteger elemsPerThread = numElements / _numberOfThreads;
    NSInteger elemsRemainder = numElements % _numberOfThreads;
    
    NSUInteger newNumElements = 0;
    NSUInteger previousThreadCountEnd = 0;
    for (NSInteger i = 0; i < _numberOfThreads; i++) {
        
        NSUInteger startElementForThread = previousThreadCountEnd;
        NSUInteger endElementForThread = startElementForThread + elemsPerThread + (i <= elemsRemainder - 1 ? 1 : 0);
        NSUInteger elemsForThread = endElementForThread - startElementForThread;
        
        // Group Varint requires special partitioning (count must be multiple of 4)
        
        if ([_encoder.encoder isKindOfClass:[GroupVarintEncoder class]]) {
            elemsForThread -= elemsForThread & 3;
        }
        
        previousThreadCountEnd = startElementForThread + elemsForThread;
        newNumElements += elemsForThread;
//        NSLog(@"[Thread #%i]: Start: %i. End: %i. Rounded element count: %i", i, startElementForThread, endElementForThread, elemsForThread);
        
        ThreadEntryCount entry;
        entry.start = startElementForThread;
        entry.end = startElementForThread + elemsForThread;
        [_threadPartitioning addObject:[NSValue valueWithBytes:&entry objCType:@encode(ThreadEntryCount)]];
    }
    
    if ([_encoder.encoder isKindOfClass:[GroupVarintEncoder class]]) {
        NSUInteger numDiscardedElements = numElements - newNumElements;
//        NSLog(@"Discarded elements: %i. To add: %i.", numDiscardedElements, (numDiscardedElements >> 2) << 2);
        if (numDiscardedElements >= 4) {
            ThreadEntryCount lastEntry;
            [_threadPartitioning[_threadPartitioning.count - 1] getValue:&lastEntry];
//            NSLog(@"Last entry before modif.: Start: %i. End: %i.", lastEntry.start, lastEntry.end);
            lastEntry.end += (numDiscardedElements >> 2) << 2;
//            NSLog(@"After modif.: End: %i.", lastEntry.end);
//            newNumElements += 4;
            _threadPartitioning[_threadPartitioning.count - 1] = [NSValue valueWithBytes:&lastEntry objCType:@encode(ThreadEntryCount)];
        }
    }
//    NSLog(@"New element count: %i. Discarded elements: %i", newNumElements, numElements - newNumElements);
    
    return _threadPartitioning;
    
}

-(NSUInteger) store:(NSString *)path withMappingAtPath:(NSString *)mappingPath inBlocks:(BOOL)blocked
{
    NSFileManager *fm = [NSFileManager defaultManager];
    
    NSUInteger blockSize = [super blockSize];
    
    BOOL fileCreated;
    if (![fm fileExistsAtPath:path])
        fileCreated = [fm createFileAtPath:path contents:nil attributes:nil];
    else
    {
        fileCreated = [fm removeItemAtPath:path error:nil];
        fileCreated = [fm createFileAtPath:path contents:nil attributes:nil];
    }
    
    BOOL mappingFileCreated;
    if (![fm fileExistsAtPath:mappingPath])
        mappingFileCreated = [fm createFileAtPath:mappingPath contents:nil attributes:nil];
    else
    {
        mappingFileCreated = [fm removeItemAtPath:mappingPath error:nil];
        mappingFileCreated = [fm createFileAtPath:mappingPath contents:nil attributes:nil];
    }
    
    if (fileCreated && mappingFileCreated)
    {
        @autoreleasepool {
            NSFileHandle *writer = [NSFileHandle fileHandleForWritingAtPath:path];
            NSFileHandle *mappingWriter = [NSFileHandle fileHandleForWritingAtPath:mappingPath];
            NSArray *sortedKeys = [[super.values allKeys] sortedArrayUsingSelector:@selector(compare:)];
            [Debug log:@"File '%@' exists. Starting write...", path];
            
            NSMutableData *toWriteData = [NSMutableData data];
            NSMutableData *mappingData = [NSMutableData data];
            NSUInteger postingListOffset = 0;
            for (NSUInteger i = 0; i < sortedKeys.count; i++)
            {
                NSString *term = sortedKeys[i];
//                path = [NSString stringWithFormat:@"/Users/lars/tmp/%@", term];
//                [fm removeItemAtPath:path error:nil];
//                [fm createFileAtPath:path contents:nil attributes:nil];
//                writer = [NSFileHandle fileHandleForWritingAtPath:path];
                
                NSData *termMappingData = [[NSString stringWithFormat:@"%08i\n", postingListOffset] dataUsingEncoding:NSASCIIStringEncoding];
                [mappingData appendData:termMappingData];
                
                NSData *prefix = [[NSString stringWithFormat:@"%@\t", term] dataUsingEncoding:NSASCIIStringEncoding];
                postingListOffset += prefix.length;
                [toWriteData appendData:prefix];
                
                NSDictionary *termDictionary = [super.values objectForKey:term];
                NSArray *sortedPostingList = [[termDictionary allKeys] sortedArrayUsingSelector:@selector(compare:)];
                [Debug log:@"Writing encoding for term '%@'. Size: %i", term, sortedPostingList.count];
                for (NSUInteger j = 0; j < sortedPostingList.count; j++)
                {
                    NSNumber *key = sortedPostingList[j];
                    NSData *data = [termDictionary objectForKey:key];
                    postingListOffset += data.length;
                    [toWriteData appendData:data];
                    if (blocked)
                    {
                        if (toWriteData.length >= blockSize)
                        {
                            [writer writeData:toWriteData];
                            [toWriteData setLength:0];
                        }
                    }
                }
                
                if (i != sortedKeys.count - 1)
                {
                    [toWriteData appendData:[@"\n" dataUsingEncoding:NSASCIIStringEncoding]];
                    postingListOffset += 1;
                }
                if (!blocked)
                {
                    [writer writeData:toWriteData];
                    [toWriteData setLength:0];
                }
            }
            [writer writeData:toWriteData];
            [writer closeFile];
            
            [mappingData appendData:[[NSString stringWithFormat:@"%08i", postingListOffset + 1] dataUsingEncoding:NSASCIIStringEncoding]];
            BOOL mappingFileCreated;
            if (![fm fileExistsAtPath:mappingPath])
                mappingFileCreated = [fm createFileAtPath:mappingPath contents:nil attributes:nil];
            else
                mappingFileCreated = YES;
            if (mappingFileCreated)
            {
                [Debug log:@"Mapping file '%@' exists. Writing mapping file...", mappingPath];
                [mappingWriter writeData:mappingData];
                [mappingWriter closeFile];
            }
        }
    }
    else
    {
        [Debug log:@"Unable to find/create file '%@'.", path];
        return -1;
    }
    
    return 0;
}

@end
