// Copyright 2013 Atbrox AS
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#import "PostingListModel.h"

#import "Debug.h"

@interface PostingListModel ()

- (void)readPostingListsBlocked;
- (void)preprocess:(PostingListEntry *)postingListEntry;
- (void)process;
- (void)postprocess:(PostingListEntry *)postingListEntry;
- (void)writePostingLists;

@end

@implementation PostingListModel
{
    NSMutableDictionary *_values;
    NSUInteger _blockSize;
    
    NSDictionary *_pairWiseTermAndByteRange;
    NSArray *_keysSortedByRange;
    
    // Semaphores
    dispatch_semaphore_t _finished;
    
    dispatch_semaphore_t _haveReadCount;
    dispatch_semaphore_t _toReadCount;
    dispatch_semaphore_t _haveProcessedCount;
    dispatch_semaphore_t _toProcessCount;
    dispatch_semaphore_t _haveWrittenCount;
    dispatch_semaphore_t _toWriteCount;
    
    // Dispatch Queues
    dispatch_queue_t finishedQueue;
    
    dispatch_queue_t preReadQueue;
    dispatch_queue_t readQueue;
    dispatch_queue_t preProcessQueue;
    dispatch_queue_t processQueue;
    dispatch_queue_t postProcessQueue;
    dispatch_queue_t writeQueue;
}

@synthesize path = _path;

+ (NSString *) STATISTIC_KEY_READ
{
    return @"read";
}

+ (NSString *) STATISTIC_KEY_PREPROCESS
{
    return @"preprocess";
}

+ (NSString *) STATISTIC_KEY_PROCESS
{
    return @"process";
}

+ (NSString *) STATISTIC_KEY_POSTPROCESS
{
    return @"postprocess";
}

+ (NSString *)STATISTIC_KEY_WRITE {
    return @"write";
}

+ (NSString *)STATISTIC_KEY_READ_WAIT {
    return @"read-wait";
}

+ (NSString *)STATISTIC_KEY_WRITE_WAIT {
    return @"write-wait";
}

+ (NSString *)STATISTIC_KEY_PROCESS_WAIT {
    return @"process-wait";
}

+ (NSString *)STATISTIC_KEY_TOTAL {
    return @"total";
}

-(id) init
{
    self = [super init];
    if (self)
    {
        _blockSize = 1;
        _values = [NSMutableDictionary dictionary];
        
        // Statistics
        _stats = [NSMutableDictionary dictionary];
        [_stats setValue:@0.0 forKey:PostingListModel.STATISTIC_KEY_READ];
        [_stats setValue:@0.0 forKey:PostingListModel.STATISTIC_KEY_PREPROCESS];
        [_stats setValue:@0.0 forKey:PostingListModel.STATISTIC_KEY_PROCESS];
        [_stats setValue:@0.0 forKey:PostingListModel.STATISTIC_KEY_POSTPROCESS];
        [_stats setValue:@0.0 forKey:PostingListModel.STATISTIC_KEY_WRITE];
        [_stats setValue:@0.0 forKey:PostingListModel.STATISTIC_KEY_READ_WAIT];
        [_stats setValue:@0.0 forKey:PostingListModel.STATISTIC_KEY_WRITE_WAIT];
        [_stats setValue:@0.0 forKey:PostingListModel.STATISTIC_KEY_PROCESS_WAIT];
        
        // Multithreading
        finishedQueue = dispatch_queue_create("Finished", NULL);
        
        preReadQueue = dispatch_queue_create("PreReadQueue", DISPATCH_QUEUE_CONCURRENT);
        readQueue = dispatch_queue_create("ReadQueue", NULL);
        preProcessQueue = dispatch_queue_create("PreProcessQueue", DISPATCH_QUEUE_CONCURRENT);
        processQueue = dispatch_queue_create("ProcessQueue", DISPATCH_QUEUE_CONCURRENT);
        postProcessQueue = dispatch_queue_create("PostProcessQueue", DISPATCH_QUEUE_CONCURRENT);
        writeQueue = dispatch_queue_create("WriteQueue", DISPATCH_QUEUE_CONCURRENT);
        
    }
    return self;
}

- (void)setPath:(NSString *)path
{
    _path = path;
    _reader = [[RawPostingListReader alloc] initFromPath:path withBlockSize:_blockSize];
}

- (NSString *)path
{
    return _path;
}

-(id) initWithPath:(NSString *)path
{
    self = [self init];
    if (self)
    {
        _path = path;
        _reader = [[RawPostingListReader alloc] initFromPath:path withBlockSize:_blockSize];
    }
    return self;
}

-(NSUInteger) blockSize
{
    return _blockSize;
}

-(void) setBlockSize:(NSUInteger)blockSize
{
    if (blockSize <= 0)
        blockSize = 1;
    if (_reader)
        _reader.blockSize = blockSize;
    _blockSize = blockSize;
}

-(void) readPostingListForTerms:(NSDictionary *)pairWiseTermAndByteRange
{
    if (_reader)
    {
        @autoreleasepool {
            _scheduledReads = [NSMutableArray array];
            _scheduledProcessing = [NSMutableArray array];
            _scheduledWrites = [NSMutableArray array];
            
            _pairWiseTermAndByteRange = pairWiseTermAndByteRange;
            [Debug log:@"Number of terms to process: %i", _pairWiseTermAndByteRange.count];
            
            _finished = dispatch_semaphore_create(0);
            
            _haveReadCount = dispatch_semaphore_create(pairWiseTermAndByteRange.count);
            _toReadCount = dispatch_semaphore_create(0);
            _haveProcessedCount = dispatch_semaphore_create(pairWiseTermAndByteRange.count);
            _toProcessCount = dispatch_semaphore_create(0);
            _haveWrittenCount = dispatch_semaphore_create(pairWiseTermAndByteRange.count);
            _toWriteCount = dispatch_semaphore_create(0);
            
            //        [self bundlePostingListReading];
            //        [self readPostingLists];
            [self _prepareWrite];
            
            NSDate *start = [NSDate date];
            [self readPostingListsBlocked];
            [self process];
            [self writePostingLists];
            
            dispatch_sync(finishedQueue, ^{
                [Debug log:@"Waiting for process..."];
                dispatch_semaphore_wait(_finished, DISPATCH_TIME_FOREVER);
                [Debug log:@"Process finished!"];
            });
            float totalDuration = [[NSDate date] timeIntervalSinceDate:start];
            [_stats setObject:@(totalDuration) forKey:[PostingListModel STATISTIC_KEY_TOTAL]];
            
            NSLog(@"Writing mapping for %i terms...", _values.count);
            for (NSString *key in _values) {
                PostingListEntry *entry = [_values objectForKey:key];
                NSLog(@"Writing mapping for entry: %@. Length: %i", entry.term, entry.dataLength);
                [self _writeMapping:entry];
            }
        }
    }
}

- (void)readPostingListsBlocked {
    dispatch_async(readQueue, ^{
        for (NSString *term in _pairWiseTermAndByteRange) {
            PostingListEntry *postingListEntry = [_values objectForKey:term];
            if (!postingListEntry) {
                postingListEntry = [[PostingListEntry alloc] init];
                postingListEntry.term = term;
                postingListEntry.range = [[_pairWiseTermAndByteRange objectForKey:term] rangeValue];
                postingListEntry.lastReadPosition = postingListEntry.range.location;
            }
            
            _currentEntry = postingListEntry;
            
            while (YES) {
                
                // Time how long this stage must wait.
                    
                    NSDate *waitStart = [NSDate date];
                    dispatch_semaphore_wait(_haveProcessedCount, DISPATCH_TIME_FOREVER);
                    float waitDuration = [[NSDate date] timeIntervalSinceDate:waitStart];
                    float waitDurationPrevious = [[_stats objectForKey:[PostingListModel STATISTIC_KEY_READ_WAIT]] floatValue];
                    [_stats setObject:[NSNumber numberWithFloat:waitDurationPrevious + waitDuration] forKey:[PostingListModel STATISTIC_KEY_READ_WAIT]];
                
                @autoreleasepool {
                    
                    NSDate *start = [NSDate date];
                    postingListEntry.buffer = [_reader getPostingList:NSMakeRange(postingListEntry.lastReadPosition, _blockSize) blockWiseRead:YES];
                    float duration = [[NSDate date] timeIntervalSinceDate:start];
                    float previousDuration = [[_stats objectForKey:[PostingListModel STATISTIC_KEY_READ]] floatValue];
                    [_stats setObject:[NSNumber numberWithFloat:previousDuration + duration] forKey:[PostingListModel STATISTIC_KEY_READ]];
                    
                    if (!postingListEntry.buffer)
                        break;
                    postingListEntry.lastReadPosition += postingListEntry.buffer.length;
                    
                    @synchronized(_scheduledProcessing) {
                        [_scheduledProcessing push:postingListEntry];
                    }
                    dispatch_semaphore_signal(_toProcessCount);
                    
                    if (_delegate)
                        [_delegate readFinished:self forPostingListEntry:postingListEntry];
                    
                    if (postingListEntry.lastReadPosition >= postingListEntry.range.location + postingListEntry.range.length) {
                        NSLog(@"Reached end at position %i. (Location: %i. Length: %i.", postingListEntry.lastReadPosition, postingListEntry.range.location, postingListEntry.range.length);
                        break;
                    }
                }
            }
            NSLog(@"[READ] Signaling outside loop. Should cascade and end.");
            dispatch_semaphore_signal(_toProcessCount);
        }
    });
}

-(void) process
{
    dispatch_async(processQueue, ^{
        [Debug log:@"Process..."];
        while (YES) {
                NSDate *waitStart = [NSDate date];
                dispatch_semaphore_wait(_toProcessCount, DISPATCH_TIME_FOREVER);
                float waitDuration = [[NSDate date] timeIntervalSinceDate:waitStart];
                float waitDurationPrevious = [[_stats objectForKey:[PostingListModel STATISTIC_KEY_PROCESS_WAIT]] floatValue];
                [_stats setObject:[NSNumber numberWithFloat:waitDurationPrevious + waitDuration] forKey:[PostingListModel STATISTIC_KEY_PROCESS_WAIT]];
            
            @autoreleasepool {
                
                PostingListEntry *postingListEntry;
                @synchronized(_scheduledProcessing)
                {
                    postingListEntry = [_scheduledProcessing pop];
                }
                if (postingListEntry == nil)
                    break;
                
                // Pre-process stage
                
                [self preprocess:postingListEntry];
                
                // Process-stage
                
                NSDate *start = [NSDate date];
                [self _process:postingListEntry];
                float duration = [[NSDate date] timeIntervalSinceDate:start];
                float previousDuration = [[_stats objectForKey:[PostingListModel STATISTIC_KEY_PROCESS]] floatValue];
                [_stats setObject:[NSNumber numberWithFloat:previousDuration + duration] forKey:[PostingListModel STATISTIC_KEY_PROCESS]];
                
                // Post-process stage
                
                [self postprocess:postingListEntry];
                
                // Signal to read another block
                
                if (_delegate)
                    [_delegate processingFinished:self forPostingListEntry:postingListEntry];
                dispatch_semaphore_signal(_haveProcessedCount);
                
                dispatch_semaphore_wait(_haveWrittenCount, DISPATCH_TIME_FOREVER);
                @synchronized(_scheduledWrites)
                {
                    [_scheduledWrites push:postingListEntry];
                }
                
                dispatch_semaphore_signal(_toWriteCount);
            }
        }
        NSLog(@"[PROC] Signaling outside loop. Should cascade and end.");
        dispatch_semaphore_signal(_toWriteCount);
    });
}

-(void) preprocess:(PostingListEntry *)postingListEntry {
//    @autoreleasepool {
        NSDate *start = [NSDate date];
        [self _preprocess:postingListEntry];
        float duration = [[NSDate date] timeIntervalSinceDate:start];
        float previousDuration = [[_stats objectForKey:[PostingListModel STATISTIC_KEY_PREPROCESS]] floatValue];
        [_stats setObject:[NSNumber numberWithFloat:previousDuration + duration] forKey:[PostingListModel STATISTIC_KEY_PREPROCESS]];
//    }
}

- (void)postprocess:(PostingListEntry *)postingListEntry {
//    @autoreleasepool {
        NSDate *start = [NSDate date];
        [self _postprocess:postingListEntry];
        float duration = [[NSDate date] timeIntervalSinceDate:start];
        float previousDuration = [[_stats objectForKey:[PostingListModel STATISTIC_KEY_POSTPROCESS]] floatValue];
        [_stats setObject:[NSNumber numberWithFloat:previousDuration + duration] forKey:[PostingListModel STATISTIC_KEY_POSTPROCESS]];
//    }
}

- (void)writePostingLists {
    
    dispatch_async(writeQueue, ^{
        [Debug log:@"Write posting list..."];
        while (YES) {
                NSDate *waitStart = [NSDate date];
                dispatch_semaphore_wait(_toWriteCount, DISPATCH_TIME_FOREVER);
                float waitDuration = [[NSDate date] timeIntervalSinceDate:waitStart];
                float waitDurationPrevious = [[_stats objectForKey:[PostingListModel STATISTIC_KEY_WRITE_WAIT]] floatValue];
                [_stats setObject:[NSNumber numberWithFloat:waitDurationPrevious + waitDuration] forKey:[PostingListModel STATISTIC_KEY_WRITE_WAIT]];
            
            @autoreleasepool {
                
                PostingListEntry *postingListEntry;
                @synchronized(_scheduledWrites) {
                    postingListEntry = [_scheduledWrites pop];
                }
                if (!postingListEntry) {
                    break;
                }
                
                NSDate *start = [NSDate date];
                BOOL writtenData = [self _writePostingList:postingListEntry];
                float duration = [[NSDate date] timeIntervalSinceDate:start];
                float previousDuration = [[_stats objectForKey:[PostingListModel STATISTIC_KEY_WRITE]] floatValue];
                [_stats setObject:[NSNumber numberWithFloat:previousDuration + duration] forKey:[PostingListModel STATISTIC_KEY_WRITE]];
                
                if (_delegate)
                    [_delegate writeFinished:self forPostingListEntry:postingListEntry];
                
                dispatch_semaphore_signal(_haveWrittenCount);
                
                if (writtenData) {
                    
                    @synchronized(_values) {
                        if (![_values objectForKey:postingListEntry.term]) {
                            [_values setValue:postingListEntry forKey:postingListEntry.term];
                        }
                    }
                }
            }
        }
        
        NSLog(@"Signaling complete processing!");
        if (_writeDataFileHandle) {
            [_writeDataFileHandle closeFile];
        }
        dispatch_semaphore_signal(_finished);
        
    });
}

-(BOOL) _preprocess:(PostingListEntry *)preProcessObject
{
    @throw [NSException exceptionWithName:NSInternalInconsistencyException
                                   reason:[NSString stringWithFormat:@"You must override %@ in a subclass", NSStringFromSelector(_cmd)]
                                 userInfo:nil];
}

-(BOOL) _process:(PostingListEntry *)processObject
{
    @throw [NSException exceptionWithName:NSInternalInconsistencyException
                                   reason:[NSString stringWithFormat:@"You must override %@ in a subclass", NSStringFromSelector(_cmd)]
                                 userInfo:nil];
}

-(BOOL) _postprocess:(PostingListEntry *)postProcessObject
{
    @throw [NSException exceptionWithName:NSInternalInconsistencyException
                                   reason:[NSString stringWithFormat:@"You must override %@ in a subclass", NSStringFromSelector(_cmd)]
                                 userInfo:nil];
}

- (BOOL)_writePostingList:(PostingListEntry *)writeProcessedObject {
    @throw [NSException exceptionWithName:NSInternalInconsistencyException
                                   reason:[NSString stringWithFormat:@"You must override %@ in a subclass", NSStringFromSelector(_cmd)]
                                 userInfo:nil];
}

- (BOOL)_writeMapping:(PostingListEntry *)writeMappingObject {
    @throw [NSException exceptionWithName:NSInternalInconsistencyException
                                   reason:[NSString stringWithFormat:@"You must override %@ in a subclass", NSStringFromSelector(_cmd)]
                                 userInfo:nil];
}

- (void)_prepareWrite {
    @throw [NSException exceptionWithName:NSInternalInconsistencyException
                                   reason:[NSString stringWithFormat:@"You must override %@ in a subclass", NSStringFromSelector(_cmd)]
                                 userInfo:nil];
}

-(NSUInteger) store:(NSString *)path withMappingAtPath:(NSString *)mappingPath inBlocks:(BOOL)blocked
{
    @throw [NSException exceptionWithName:NSInternalInconsistencyException
                                   reason:[NSString stringWithFormat:@"You must override %@ in a subclass", NSStringFromSelector(_cmd)]
                                 userInfo:nil];
}


@end
