// Copyright 2013 Atbrox AS
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#import "PostingListCompression.h"

@implementation PostingListCompression
{
    NSString *pathToUnencodedData, *pathToEncodedData;
    
    NSMutableDictionary *toEncode;
    NSMutableDictionary *toDecode;
    
    NSMutableArray *encodeIterations;
    NSMutableArray *decodeIterations;
}

@synthesize encodingModel = _encodingModel, decodingModel = _decodingModel;

+(NSRange) postingListRange:(NSString *)path forTerm:(NSString *)term atLineNo:(NSUInteger)lineNo withFixedByteSize:(NSUInteger)byteSize
{
    NSUInteger readLocation = (lineNo - 1) * (byteSize + 1);
    NSFileHandle *fileMappingHandle = [NSFileHandle fileHandleForReadingAtPath:path];
    
    [fileMappingHandle seekToFileOffset:readLocation];
    NSData *listStart = [fileMappingHandle readDataOfLength:byteSize];

    [fileMappingHandle seekToFileOffset:readLocation + byteSize + 1];
    NSData *listEnd = [fileMappingHandle readDataOfLength:byteSize];
    
    [fileMappingHandle closeFile];
    
    NSUInteger start = [[[NSString alloc] initWithData:listStart encoding:NSASCIIStringEncoding] integerValue] + [term dataUsingEncoding:NSUTF8StringEncoding].length + 1;
    NSUInteger end = [[[NSString alloc] initWithData:listEnd encoding:NSASCIIStringEncoding] integerValue];
    
    return NSMakeRange(start, end - start - 1);
}

+(NSUInteger) numberOfTerms:(NSString *)path withFixedByteSize:(NSUInteger)byteSize
{
    NSFileHandle *fileMappingHandle = [NSFileHandle fileHandleForReadingAtPath:path];
    if (fileMappingHandle)
    {
        NSUInteger size = [fileMappingHandle seekToEndOfFile];
        return size / (byteSize + 1) - 1;
    }
    return 0;
}

-(id) init
{
    if (self = [super init])
    {
        _encodingModel = [[ParallelEncodingModel alloc] init];
        _decodingModel = [[ParallelDecodingModel alloc] init];
        
        _iterations = 1;
        
        toEncode = [NSMutableDictionary dictionary];
        toDecode = [NSMutableDictionary dictionary];
    }
    return self;
}

-(void) setPathToUnencodedData:(NSString *)path
{
    pathToUnencodedData = path;
    _encodingModel.path = pathToUnencodedData;
}

-(NSString *) pathToUnencodedData
{
    return pathToUnencodedData;
}

-(void) setPathToEncodedData:(NSString *)path
{
    pathToEncodedData = path;
    _decodingModel.path = pathToEncodedData;
}

-(NSString *) pathToEncodedData
{
    return pathToEncodedData;
}

-(void) addTermForDecode:(NSString *)term inRange:(NSRange)range
{
    [toDecode setObject:[NSValue valueWithRange:range] forKey:term];
}

-(void) addTermForEncode:(NSString *)term inRange:(NSRange)range
{
    [toEncode setObject:[NSValue valueWithRange:range] forKey:term];
}

- (void) decodeUsing:(id<DecoderProtocol>)decoder
{
    decodeIterations = [NSMutableArray array];
    [_decodingModel setBlockSize:_blockSize];
    _decodingModel.decoder.decoder = decoder;
    for (NSUInteger i = 0; i < _iterations; i++) {
        [_decodingModel readPostingListForTerms:toDecode];
        if (_delegate) {
            [_delegate iterationFinished:i operation:@"decode" values:_decodingModel.values statistics:_decodingModel.stats];
        }
        [decodeIterations addObject:_decodingModel.stats];
    }
}

- (void) encodeUsing:(id<EncoderProtocol>)encoder
{
    encodeIterations = [NSMutableArray array];
    [_encodingModel setBlockSize:_blockSize];
    _encodingModel.encoder.encoder = encoder;
    for (NSUInteger i = 0; i < _iterations; i++) {
        [_encodingModel readPostingListForTerms:toEncode];
        if (_delegate) {
            [_delegate iterationFinished:i operation:@"encode" values:_encodingModel.values statistics:_encodingModel.stats];
        }
        [encodeIterations addObject:_encodingModel.stats];
    }
}

-(void) storeDecodedData:(NSString *)path withMappingAtPath:(NSString *)mappingPath inBlocks:(BOOL)blocked
{
    [_decodingModel store:path withMappingAtPath:mappingPath inBlocks:blocked];
}

-(void) storeEncodedData:(NSString *)path withMappingAtPath:(NSString *)mappingPath inBlocks:(BOOL)blocked
{
    [_encodingModel store:path withMappingAtPath:mappingPath inBlocks:blocked];
}

@end
