// Copyright 2013 Atbrox AS
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#import "PostingListEncoder.h"

#import "Debug.h"
#import "ParallelEncodingModel.h"

@interface PostingListEncoder ()

- (NSMutableData *)sortDictionaryByKeys:(NSDictionary *)dictionary;

@end

@implementation PostingListEncoder

-(id) initWithInput:(CFMutableArrayRef *)array
{
    self = [super init];
    if (self)
        _input = *array;
    
    return self;
}

- (NSMutableData *)sortDictionaryByKeys:(NSDictionary *)dictionary {
    NSMutableArray *sortedArray = [NSMutableArray arrayWithArray:[dictionary allKeys]];
    NSMutableData *sortedData = [NSMutableData data];
    
    for (NSNumber *key in [dictionary allKeys])
        [sortedArray setObject:[dictionary objectForKey:key] atIndexedSubscript:[key unsignedIntegerValue]];
    for (NSData *data in sortedArray) {
        [sortedData appendData:data];
    }
    return sortedData;
}

-(NSData *) encode:(NSArray *)threadPartitioning
{
    if (_input)
    {
        @autoreleasepool {
            NSMutableData *encodedData = [NSMutableData data];
            NSUInteger itemCount = CFArrayGetCount(_input);
            [Debug log:@"Processing list with %i items", itemCount];
            NSDate *startTime = [NSDate date];
            if (threadPartitioning)
            {
                // Thread setup
                NSMutableDictionary *dictionaryWithThreadData = [NSMutableDictionary dictionary];
                dispatch_queue_t encodeQueue = dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_HIGH, 0);
                dispatch_group_t encodeGroup = dispatch_group_create();
                
                for (NSInteger i = 0; i < threadPartitioning.count; i++)
                {
                    dispatch_group_async(encodeGroup, encodeQueue, ^{
                        
                        ThreadEntryCount entry;
                        [threadPartitioning[i] getValue:&entry];
                        // Add remainder to start if thread # is above 0
                        NSUInteger start = entry.start;
                        // Add 1 element to each thread to divide workload across all
                        NSUInteger end = entry.end;
                        [Debug log:@"[Thread #%i] Processing element [%i to %i>...", i, start, end];
                        
                        @autoreleasepool {
//                            NSMutableData *threadData = [NSMutableData dataWithCapacity:end - start];
                            NSUInteger threadValueCount = end - start;
                            NSUInteger *threadValues = (NSUInteger *)malloc(sizeof(NSUInteger) * threadValueCount);
                            CFArrayGetValues(_input, CFRangeMake(start, threadValueCount), (const void **)threadValues);
                            CFArrayRef threadArray = CFArrayCreate(kCFAllocatorDefault, (const void **)threadValues, threadValueCount, NULL);
//                            NSLog(@"[Thread #%i] Encoding: %@", i, threadArray);
                            NSData *threadData = [_encoder encode:threadArray];
                            CFRelease(threadArray);
                            free(threadValues);
//                            for (NSUInteger j = start; j < end; j++)
//                            {
//                                NSData *encodedValue = [_encoder encodeNumber:[_input objectAtIndex:j]];
//                                //                            NSLog(@"[Thread #%i] ENCODED Current: %i. Input count: %i", i, j, _input.count);
//                                [threadData appendData:encodedValue];
//                            }
                            @synchronized(self) {
                                [dictionaryWithThreadData setObject:threadData forKey:@(i)];
                            }
                        }
                        //                    NSLog(@"[Thread #%li] Encoded list as seen from thread:\n%@", i, encodedPostingList.debugDescription);
                    });
                }
                [Debug log:@"Waiting for threads to finish..."];
                dispatch_group_wait(encodeGroup, DISPATCH_TIME_FOREVER);
                //            NSLog(@"Encoded list as seen MAIN thread:\n%@", encodedPostingList.debugDescription);
                [encodedData appendData:[self sortDictionaryByKeys:dictionaryWithThreadData]];
            }
            else
            {
                [Debug log:@"Executing using main thread. Processing element [0 to %i>...", itemCount];
                [encodedData appendData:[_encoder encode:_input]];
                    //                NSLog(@"Input value: %@. Encoded value: %@. Inserted at index: %li", input[i], encodedValue.debugDescription, i);
            }
            NSDate *endTime = [NSDate date];
            [Debug log:@"Finished encoding! Took: %f.", [endTime timeIntervalSinceDate:startTime]];
            return encodedData;
        }
    }
    return nil;
}

- (NSUInteger)encodeLength {
    return _encoder.encodedDataLength;
}

@end
