// Copyright 2013 Atbrox AS
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#import "RawPostingListReader.h"

#import "Debug.h"

#define TERMID_SIZE 8
#define LIST_LEN_SIZE 8
#define LIST_MAX_SIZE 64*1024*1024


@implementation RawPostingListReader
{
    NSFileManager *fileManager;
    NSFileHandle *fileHandler;
}

-(id) init
{
    self = [super init];
    if (self)
    {
        fileManager = [NSFileManager defaultManager];
    }
    return self;
}

-(id) initFromPath:(NSString *)path withBlockSize:(NSUInteger)blockSize
{
    self = [self init];
    if (self)
    {
        if ([fileManager fileExistsAtPath:path])
        {
            _blockSize = blockSize;
            [self loadFromPath:path];
        }
    }
    return self;
}

-(id) initFromFileHandle:(NSFileHandle *)fileHandle withBlockSize:(NSUInteger)blockSize
{
    self = [self init];
    if (self)
    {
        fileHandler = fileHandle;
        _length = [fileHandler seekToEndOfFile];
        [fileHandler seekToFileOffset:0];
    }
    return self;
}

-(void) loadFromPath:(NSString *)path
{
    [Debug log:@"Attempting to load file from %@", path];
    if ([fileManager fileExistsAtPath:path])
    {
        [Debug log:@"File exists! Continuing..."];
        fileHandler = [NSFileHandle fileHandleForReadingAtPath:path];
        _path = path;
        _length = [fileHandler seekToEndOfFile];
        [fileHandler seekToFileOffset:0];
    }
}

-(void) closeFile
{
    if (fileHandler)
        [fileHandler closeFile];
}

-(NSData *) readData
{
    return [self readData:_blockSize];
}

-(NSData *) readData:(NSUInteger)length
{
    if (fileHandler)
        return [fileHandler readDataOfLength:length];
    return nil;
}

-(NSData *) getPostingList:(NSRange)byteRange blockWiseRead:(BOOL)blockWise
{
    if (fileHandler)
    {
        [fileHandler seekToFileOffset:byteRange.location];
        [Debug log:@"Reading from position %lli", [fileHandler offsetInFile]];
        [Debug log:@"Read list length as %i bytes. Block size: %i.", byteRange.length, _blockSize];
        NSData *readData;
        NSDate *start = [NSDate date];
        if (blockWise)
        {
            NSUInteger blocksToRead = ceil(byteRange.length / (double)_blockSize);
            [Debug log:@"Reading %i blocks (total: %i bytes. File length: %i bytes)", blocksToRead, blocksToRead * _blockSize, _length];
            readData = [NSData dataWithData:[fileHandler readDataOfLength:MIN(blocksToRead * _blockSize, _length)]];
        }
        else
            readData = [NSData dataWithData:[fileHandler readDataOfLength:MIN(byteRange.length, _length)]];

        [Debug log:@"Read %i bytes. Took: %.4fs.", readData.length, [[NSDate date] timeIntervalSinceDate:start]];
        return readData;
    }
    return nil;
}

@end
