// Copyright 2013 Atbrox AS
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#import "FileReader.h"

#include <sys/mman.h>

#define FLUSH_SIZE 512*1024*1024
#define FLUSH_BLOCK_SIZE 256*1024

@interface FileReader ()

-(void) stream:(NSMutableString *)readLog;
-(void) randomAccess:(NSMutableString *)readLog;

@end

@implementation FileReader
{
    NSFileManager *_fileManager;
    NSUInteger _flushSize;
    NSString *_tmpInput;
}

@synthesize blockSize = _blockSize, iterations = _iterations, currentIteration = _currentIteration, numberOfBlocks = _numberOfBlocks;
@synthesize input = _input, log = _log, totalDuration = _totalDuration, averageDuration = _averageDuration;
@synthesize averageTransferRate = _averageTransferRate, totalBytes = _totalBytes, currentReadBytes = _currentReadBytes;
@synthesize start = _start, end = _end;

-(id) init
{
    self = [super init];
    if (self)
    {
        _fileManager = [NSFileManager defaultManager];
        _blockSize = 1;
        _iterations = 1;
    }
    return self;
}

-(id) initFromPath:(NSString *)path
{
    self = [self init];
    if (self)
    {
        if ([_fileManager fileExistsAtPath:path])
        {
            NSLog(@"Found file at path '%@'", path);
            _input = path;
        }
        else
            NSLog(@"No such file: '%@'", path);
    }
    return self;
}

-(void) resetCounters
{
    _currentReadBytes = 0;
    _currentIteration = 0;
    _totalDuration = 0.0;
}

-(void) read:(BOOL)random
{
    if (_input) {
        if ([_fileManager fileExistsAtPath:_input])
        {
            _totalBytes = [[[_fileManager attributesOfItemAtPath:_input error:nil] objectForKey:NSFileSize] integerValue];
            _tmpInput = [NSTemporaryDirectory() stringByAppendingPathComponent:@"data.random"];
            [_fileManager copyItemAtPath:_input toPath:_tmpInput error:nil];
        }
        else
        {
            NSLog(@"No such file: '%@'", _tmpInput);
            return;
        }
    }
    else
    {
        NSLog(@"File path is not set.");
        return;
    }

        
    _numberOfBlocks = ceil(_totalBytes / (double)_blockSize);
    NSLog(@"Reading is started. Random: %i. Total file size: %i. Block size: %i. Number of blocks: %i. Iterations: %i", random, _totalBytes, _blockSize, _numberOfBlocks, _iterations);
    NSMutableString *readLog = [[NSMutableString alloc] init];
    [self resetCounters];
    _start = [NSDate date];
    
    if (random)
        [self randomAccess:readLog];
    else
        [self stream:readLog];
    
//    NSString *lastPath = [NSString stringWithFormat:@"%@.%i", _input, _currentIteration - 1];
//    NSLog(@"Benchmark done. Moving file %@ to %@", lastPath, _input);
//    [_fileManager moveItemAtPath:lastPath toPath:_input error:nil];
    NSLog(@"Doing cleanup...");
    NSArray *tempFiles = [_fileManager contentsOfDirectoryAtPath:NSTemporaryDirectory() error:nil];
    BOOL isDirectory;
    for (NSString *file in tempFiles)
    {
        NSLog(@"Found temporary file '%@'...", file);
        NSString *completePath = [NSTemporaryDirectory() stringByAppendingPathComponent:file];
        if ([_fileManager fileExistsAtPath:completePath isDirectory:&isDirectory] && !isDirectory) {
            NSString *subString = [file substringToIndex:11];
            NSLog(@"Substring: %@", subString);
            if ([[file substringToIndex:11] isEqualToString:@"data.random"])
            {
                NSLog(@"Removing file: '%@'...", completePath);
                [_fileManager removeItemAtPath:completePath error:nil];
            }
        }
    }
    
    _end = [NSDate date];
    _averageDuration = _totalDuration / _iterations;
    _averageTransferRate = (_totalBytes / (1024*1024)) / _averageDuration;
    
    _log = [NSMutableString stringWithFormat:@"Test start: %@\nTest end: %@.\nWas Random: %i\nIterations: %i.\nTotal duration: %.2f s.\nAverage iteration duration: %.2f s.\nAverage transfer rate: %.2f MB/s\n\n",
            _start.description, _end.description, random, _iterations, _totalDuration, _averageDuration, _averageTransferRate];
    NSLog(@"%@", _log);
    [_log appendString:readLog];
}

-(void) stream:(NSMutableString *)readLog
{
    while (_currentIteration < _iterations) {
        NSString *newPath = [NSString stringWithFormat:@"%@.%i", _tmpInput, _currentIteration];
        [_fileManager copyItemAtPath:_tmpInput toPath:newPath error:nil];
        
        if ([_fileManager fileExistsAtPath:newPath])
        {
            NSLog(@"Reading from '%@'...", newPath);
            NSFileHandle *fileHandle = [NSFileHandle fileHandleForReadingAtPath:newPath];
            if (fileHandle)
            {
                
                @synchronized(self)
                {
                    _currentReadBytes = 0;
                }
                @autoreleasepool {
                    double totalDurationPerIteration = 0.0;
                    while (_currentReadBytes < _totalBytes) {
                        @autoreleasepool {
                            NSDate *subStart = [NSDate date];
                            
                            NSData *buffer = [fileHandle readDataOfLength:_blockSize];
                            
                            float duration = [[NSDate date] timeIntervalSinceDate:subStart];
                            totalDurationPerIteration += duration;
                            _currentReadBytes += buffer.length;
                        }
                    }
                    
                    float averageReadPerBlock = (totalDurationPerIteration / _numberOfBlocks) * 1000;
                    float averageTransferRate = (_totalBytes / (1024 * 1024) / totalDurationPerIteration);
                    NSString *logEntry = [NSString stringWithFormat:@"Iteration #%i. Took: %f s. Average transfer duration per block: %.2f ms. Transfer rate: %f. MB/s",
                                          _currentIteration, totalDurationPerIteration, averageReadPerBlock, averageTransferRate];
                    NSLog(@"%@", logEntry);
                    [readLog appendString:[NSString stringWithFormat:@"%@\n", logEntry]];
                    _totalDuration += totalDurationPerIteration;
                    _currentIteration++;
                }
                [fileHandle closeFile];
                [_fileManager removeItemAtPath:newPath error:nil];
            }
            else
            {
                NSLog(@"An error occured:");
                perror("fopen");
                break;
            }
        }
        else
        {
            NSLog(@"No such file: '%@'!", newPath);
            _currentIteration++;
        }
    }
}

-(void) randomAccess:(NSMutableString *)readLog
{
    NSMutableArray *offsets = [NSMutableArray arrayWithCapacity:_numberOfBlocks];
    for (NSUInteger i = 0; i < _totalBytes; i += _blockSize)
        [offsets addObject:[NSNumber numberWithInteger:i]];
    
    while (_currentIteration < _iterations) {
        NSString *newPath = [NSString stringWithFormat:@"%@.%i", _tmpInput, _currentIteration];
        [_fileManager copyItemAtPath:_tmpInput toPath:newPath error:nil];
        
        if ([_fileManager fileExistsAtPath:newPath])
        {
            NSLog(@"Reading from '%@'...", newPath);
            NSFileHandle *fileHandle = [NSFileHandle fileHandleForReadingAtPath:newPath];
            if (fileHandle)
            {
                
                @synchronized(self)
                {
                    _currentReadBytes = 0;
                }
                @autoreleasepool {
                    
                    NSUInteger count = [offsets count];
                    for (NSUInteger i = 0; i < count; ++i)
                    {
                        NSInteger nElements = count - i;
                        NSInteger n = (arc4random() % nElements) + i;
                        [offsets exchangeObjectAtIndex:i withObjectAtIndex:n];
                    }
                    
                    double totalDurationPerIteration = 0.0;
                    NSUInteger offsetDelta = 0;
                    
                    for (NSNumber *offset in offsets) {
                        @autoreleasepool {
                            NSDate *subStart = [NSDate date];
                            
                            NSUInteger currentPosition = [fileHandle offsetInFile];
                            [fileHandle seekToFileOffset:[offset integerValue]];
                            offsetDelta += ABS(currentPosition - [fileHandle offsetInFile]);
                            NSData *buffer = [fileHandle readDataOfLength:_blockSize];
                            
                            float duration = [[NSDate date] timeIntervalSinceDate:subStart];
                            totalDurationPerIteration += duration;
                            _currentReadBytes += buffer.length;
                        }
                    }
                    
                    float averageReadPerBlock = (totalDurationPerIteration / _numberOfBlocks) * 1000;
                    float averageTransferRate = (_totalBytes / (1024 * 1024) / totalDurationPerIteration);
                    NSUInteger averageOffsetDelta = offsetDelta / offsets.count;
                    NSString *logEntry = [NSString stringWithFormat:@"Iteration #%i. Took: %f s. Average offset distance: %i. Average transfer duration per block: %.2f ms. Transfer rate: %f. MB/s",
                                          _currentIteration, totalDurationPerIteration, averageOffsetDelta, averageReadPerBlock, averageTransferRate];
                    NSLog(@"%@", logEntry);
                    [readLog appendString:[NSString stringWithFormat:@"%@\n", logEntry]];
                    _totalDuration += totalDurationPerIteration;
                    _currentIteration++;
                }
                [fileHandle closeFile];
                [_fileManager removeItemAtPath:newPath error:nil];
            }
            else
            {
                NSLog(@"An error occured:");
                perror("fopen");
                break;
            }
        }
        else
        {
            NSLog(@"No such file: '%@'!", newPath);
            _currentIteration++;
        }
    }
}

-(void) writeLog:(NSString *)path
{
    if (_log)
    {
        NSLog(@"Writing output from read session to '%@'", path);
        if (![_fileManager fileExistsAtPath:path])
            [_fileManager createFileAtPath:path contents:nil attributes:nil];
        NSFileHandle *saveHandle = [NSFileHandle fileHandleForWritingAtPath:path];
        [saveHandle writeData:[_log dataUsingEncoding:NSUTF8StringEncoding]];
        [saveHandle closeFile];
    }
}

@end
